#pragma once

#ifndef BLE_COMPONENT_H_
#define BLE_COMPONENT_H_

/**
 * @Brief BLE Component library, this will use the BLE service and handle the data. This will be the source of the truth.
 * @Updated on 08-04-2023
 **/

// Std Libs
#include <stdbool.h>
#include <stdint.h>
#include <string.h>

// NRF modules
#include "nrf.h"
#include "nrf_ble_gatt.h"
#include "nrf_ble_scan.h"
#include "nrf_power.h"
#include "nrf_pwr_mgmt.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_sdh_soc.h"

// BLE modules
#include "ble.h"
#include "ble_advdata.h"
#include "ble_advertising.h"
#include "ble_conn_state.h"
#include "ble_gap.h"
#include "ble_srv_common.h"

// Platform Libs
#include "app_error.h"
#include "app_timer.h"
#include "nordic_common.h"

// others
#include "gatt_service.h"

#define EDDY_STONE_UUID_HSB 170
#define EDDY_STONE_UUID_LSB 254

#define APP_BLE_CONN_CFG_TAG 1 /**< A tag identifying the SoftDevice BLE configuration. */

#define APP_ADV_DURATION 100 /**< The advertising duration in units of 10 milliseconds. */

#define HUB_BEACON_MAC_ADDRESS_LEN (sizeof(uint8_t) * MAC_ADDRESS_SIZE)
#define HUB_BEACON_INSTANCE_ID_LEN (sizeof(uint8_t) * INSTANCE_ID_SIZE)
#define HUB_BEACON_NAMESPACE_LEN (sizeof(uint8_t) * NAMESPACE_SIZE)
#define HUB_BEACON_UUID_LEN (sizeof(uint8_t) * UUID_SIZE)

/**
 * Data Strucuture to capture the eddy stone data
 **/
typedef struct eddy_stone_adv_data_t {
  uint8_t type;
  uint8_t uuid[UUID_SIZE];
  uint8_t rx_power;
  uint8_t namespace_id[NAMESPACE_SIZE];
  uint8_t instance_id[INSTANCE_ID_SIZE];
  uint8_t rssi;
  uint8_t address[MAC_ADDRESS_SIZE];
} eddy_stone_adv_data;

/** @brief This function is to intialize the Hub scanning
 *
 **/
void hub_beacon_scan_init();

/**
 * This method is used to start scan for hub beacons
 **/
void start_hub_beacon_scan(void);

/**
 * This method is used to stop scan for hub beacons
 **/
void stop_hub_beacon_scan(void);

/**
 * Function to initialise the GATT modules
 **/
void gatt_module_init(const ble_evt_handler_t ble_handler, const ble_adv_evt_handler_t ble_adv_evt_handler, ble_gap_addr_t *tag_mac_addr, uint32_t *tag_start_epoch);

/**
 * This method is used to start the GATT server for the data transfer
 **/
void start_gatt_server();

/**
 *  This method is used to stop the GATT server
 **/
void stop_gatt_server();

/**@brief Function for initializing advertising.
 */
void gatts_advertising_init(const ble_adv_evt_handler_t ble_adv_evt_handler);

/**@brief Function for starting advertising so that hub can connect for data tranfer.
 */
void gatts_advertising_start(void);

/**@brief Function to know when the BLE scan was stopped
 */
uint64_t get_scan_finshed_epoch();

bool is_data_transfer_in_progress();

#endif // BLE_COMPONENT_H_