#pragma once
#ifndef BLE_SERVICE_H__
#define BLE_SERVICE_H__

/**
* @Brief BLE service library, this will have only the BLE realted functions
* @Updated on 08-04-2023 
**/

//Std Libs
#include <stdbool.h>
#include <stdint.h>

//BLE Libs
#include "ble.h"
#include "ble_srv_common.h"

//Custom Libs
#include "tag_utils.h"

/**@brief   Macro for defining a ble_hrs instance.
 *
 * @param   _name   Name of the instance.
 * @hideinitializer
 */
#define BLE_CUS_DEF(_name)          \
  static ble_cus_t _name;           \
  NRF_SDH_BLE_OBSERVER(_name##_obs, \
      BLE_HRS_BLE_OBSERVER_PRIO,    \
      ble_cus_on_ble_evt, &_name)

// This is the BASE UUID: 00000000-0000-1000-8000-00805F9B34FB
#define BLE_SERVICE_UUID_BASE \
  { 0xFB, 0x34, 0x9B, 0x5F, 0x80, 0x00, 0x00, 0x80, 0x00, 0x10, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }

#define DATA_TRANSFER_SERVICE_UUID 0X010B
#define DATA_TRANFER_CHAR_UUID 0x1401

#define DATA_PACKET_LENGTH NRF_SDH_BLE_GATT_MAX_MTU_SIZE

/**@brief Custom Service event type. */
typedef enum {
  BLE_CUS_EVT_NOTIFICATION_ENABLED,  /**< Custom value notification enabled event. */
  BLE_CUS_EVT_NOTIFICATION_DISABLED, /**< Custom value notification disabled event. */
  BLE_CUS_EVT_DISCONNECTED,
  BLE_CUS_EVT_CONNECTED
} ble_cus_evt_type_t;
/**@brief Custom Service event. */

typedef struct
{
  ble_cus_evt_type_t evt_type; /**< Type of event. */
} ble_cus_evt_t;

/**
 * Data Strucuture to capture the tag data
 **/
typedef struct tag_data_t {
  uint8_t battery_percentage;
  uint8_t rssi;
  uint8_t address[6];
  uint32_t contact_mS;
} tag_data;

// Forward declaration of the ble_cus_t type.
typedef struct ble_cus_s ble_cus_t;

/**@brief Custom Service event handler type. */
typedef void (*ble_cus_evt_handler_t)(ble_cus_t *p_bas, ble_cus_evt_t *p_evt);
/**@brief Battery Service init structure. This contains all options and data needed for
 *        initialization of the service.*/

typedef struct
{
  ble_cus_evt_handler_t evt_handler;                      /**< Event handler to be called for handling events in the Custom Service. */
  uint8_t initial_custom_value;                           /**< Initial custom value */
  ble_srv_cccd_security_mode_t custom_value_char_attr_md; /**< Initial security level for Custom characteristics attribute */
} ble_cus_init_t;

/**@brief Custom Service structure. This contains various status information for the service. */
struct ble_cus_s {
  ble_cus_evt_handler_t evt_handler;             /**< Event handler to be called for handling events in the Custom Service. */
  uint16_t service_handle;                       /**< Handle of Custom Service (as provided by the BLE stack). */
  ble_gatts_char_handles_t custom_value_handles; /**< Handles related to the Custom Value characteristic. */
  uint16_t conn_handle;                          /**< Handle of the current connection (as provided by the BLE stack, is BLE_CONN_HANDLE_INVALID if not in a connection). */
  uint8_t uuid_type;
};

/**
 * @brief Function for intializing the tag Ble services
 */
uint32_t ble_tag_services_init(ble_cus_t *p_cus, const ble_cus_init_t *p_cus_init, tag_data *tag_data_array_p, int *current_pointer_index);

/**@brief Function for handling the Application's BLE Stack events.
 *
 * @details Handles all events from the BLE stack of interest to the Battery Service.
 *
 * @note
 *
 * @param[in]   p_cus      Custom Service structure.
 * @param[in]   p_ble_evt  Event received from the BLE stack.
 */
void ble_cus_on_ble_evt(ble_evt_t const *p_ble_evt, void *p_context);
#endif // BLE_SERVICE_H__