#include "gatt_service.h"

/**@brief   Macro for defining a ble_hrs instance.
 *
 * @param   _name   Name of the instance.
 * @hideinitializer
 */

static on_indicate_handler_t indicate_handler_;

static on_write_handler_t write_handler_;

static on_read_handler_t read_handler_;

static bool is_client_connected_;

/**@brief Function for handling the Connect event.
 *
 * @param[in]   p_cus       Custom Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_connect(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt) {
    sd_ble_gap_adv_stop(NULL);
  p_cus->conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
  p_cus->evt_handler(p_cus, p_ble_evt);
}

/**@brief Function for handling the Disconnect event.
 *
 * @param[in]   p_cus       Custom Service structure.
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 */
static void on_disconnect(ble_cus_t *p_cus, ble_evt_t const *p_ble_evt) {
  UNUSED_PARAMETER(p_ble_evt);
  p_cus->conn_handle = BLE_CONN_HANDLE_INVALID;
  p_cus->evt_handler(p_cus, p_ble_evt);
  stop_gatt_service();
}

/**@brief Function used to add the contact epoch characteristic to the tag service
 */
static uint32_t tag_add_contact_epoch_characteristic(ble_cus_t *p_cus, const ble_cus_init_t *p_cus_init) {
  uint32_t err_code;
  ble_gatts_char_md_t char_md;
  ble_gatts_attr_md_t cccd_md;
  ble_gatts_attr_t attr_char_value;
  ble_uuid_t ble_uuid;
  ble_gatts_attr_md_t attr_md;

  // Add Custom Value characteristic
  memset(&cccd_md, 0, sizeof(cccd_md));

  //  Read  operation on cccd should be possible without authentication.
  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);

  cccd_md.write_perm = p_cus_init->custom_value_char_attr_md.cccd_write_perm;
  cccd_md.vloc = BLE_GATTS_VLOC_STACK;

  memset(&char_md, 0, sizeof(char_md));

  char_md.char_props.read = 1;
  char_md.char_props.write = 0;
  char_md.char_props.indicate = 0;
  char_md.p_char_user_desc = NULL;
  char_md.p_char_pf = NULL;
  char_md.p_user_desc_md = NULL;
  char_md.p_cccd_md = &cccd_md;
  char_md.p_sccd_md = NULL;

  ble_uuid.type = p_cus->uuid_type;
  ble_uuid.uuid = CONTACT_EPOCH_CHAR_UUID;

  memset(&attr_md, 0, sizeof(attr_md));

  attr_md.read_perm = p_cus_init->custom_value_char_attr_md.read_perm;
  attr_md.write_perm = p_cus_init->custom_value_char_attr_md.write_perm;
  attr_md.vloc = BLE_GATTS_VLOC_STACK;
  attr_md.rd_auth = 0;
  attr_md.wr_auth = 0;
  attr_md.vlen = 0;

  memset(&attr_char_value, 0, sizeof(attr_char_value));

  attr_char_value.p_uuid = &ble_uuid;
  attr_char_value.p_attr_md = &attr_md;
  attr_char_value.init_offs = 0;
  attr_char_value.max_len = sizeof(uint32_t);

  err_code = sd_ble_gatts_characteristic_add(p_cus->service_handle, &char_md,
      &attr_char_value,
      &p_cus->custom_value_handles);
  if (err_code != NRF_SUCCESS) {
    NRF_LOG_ERROR("GATT_SERVICE-- Error adding the contact epoch Char");

    return err_code;
  } // endif
  NRF_LOG_DEBUG("GATT_SERVICE-- Successfully added the  contact epoch Char");

  return NRF_SUCCESS;
}

/**@brief Function used to add the Data characteristic to the tag service
 */
static uint32_t tag_add_data_characteristic(ble_cus_t *p_cus, const ble_cus_init_t *p_cus_init) {
  uint32_t err_code;
  ble_gatts_char_md_t char_md;
  ble_gatts_attr_md_t cccd_md;
  ble_gatts_attr_t attr_char_value;
  ble_uuid_t ble_uuid;
  ble_gatts_attr_md_t attr_md;

  // Add Custom Value characteristic
  memset(&cccd_md, 0, sizeof(cccd_md));

  //  Read  operation on cccd should be possible without authentication.
  BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);

  cccd_md.write_perm = p_cus_init->custom_value_char_attr_md.cccd_write_perm;
  cccd_md.vloc = BLE_GATTS_VLOC_STACK;

  memset(&char_md, 0, sizeof(char_md));

  char_md.char_props.read = 0;
  char_md.char_props.write_wo_resp = 1;
  char_md.char_props.indicate = 1;
  char_md.p_char_user_desc = NULL;
  char_md.p_char_pf = NULL;
  char_md.p_user_desc_md = NULL;
  char_md.p_cccd_md = &cccd_md;
  char_md.p_sccd_md = NULL;

  ble_uuid.type = p_cus->uuid_type;
  ble_uuid.uuid = DATA_TRANFER_CHAR_UUID;

  memset(&attr_md, 0, sizeof(attr_md));

  attr_md.read_perm = p_cus_init->custom_value_char_attr_md.read_perm;
  attr_md.write_perm = p_cus_init->custom_value_char_attr_md.write_perm;
  attr_md.vloc = BLE_GATTS_VLOC_STACK;
  attr_md.rd_auth = 0;
  attr_md.wr_auth = 0;
  attr_md.vlen = 0;

  memset(&attr_char_value, 0, sizeof(attr_char_value));

  attr_char_value.p_uuid = &ble_uuid;
  attr_char_value.p_attr_md = &attr_md;
  attr_char_value.init_offs = 0;
  attr_char_value.max_len = sizeof(char) * DATA_PACKET_LENGTH;

  err_code = sd_ble_gatts_characteristic_add(p_cus->service_handle, &char_md,
      &attr_char_value,
      &p_cus->custom_value_handles);
  if (err_code != NRF_SUCCESS) {
    NRF_LOG_ERROR("GATT_SERVICE-- Error addeding the Data transfer Char");

    return err_code;
  } // endif
  NRF_LOG_DEBUG("GATT_SERVICE-- Successfully added the Data transfer Char");
  return NRF_SUCCESS;
  // return tag_add_contact_epoch_characteristic(p_cus, p_cus_init);
}

void ble_events_handler(ble_evt_t const *p_ble_evt, void *p_context) {
  ble_cus_t *p_cus = (ble_cus_t *)p_context;

  NRF_LOG_DEBUG("GATT_SERVICE-- BLE event received. Event type = %d\r\n", p_ble_evt->header.evt_id);
  if (p_cus == NULL || p_ble_evt == NULL) {
    return;
  }

  switch (p_ble_evt->header.evt_id) {
  case BLE_GAP_EVT_CONNECTED: {
    NRF_LOG_DEBUG("GATT_SERVICE-- Connect EVent");
    is_client_connected_ = true;
    on_connect(p_cus, p_ble_evt);
    break;
  }
  case BLE_GAP_EVT_DISCONNECTED: {
    NRF_LOG_DEBUG("GATT_SERVICE-- DisConnect EVent");
    on_disconnect(p_cus, p_ble_evt);
    break;
  }
  case BLE_GATTS_EVT_WRITE:
    NRF_LOG_DEBUG("GATT_SERVICE-- Write EVent");
    if (write_handler_ != NULL) {
      write_handler_(p_cus, p_ble_evt);
    } else {
      NRF_LOG_ERROR("GATT_SERVICE-- Write handler was NULL");
    }
    break;

  case BLE_GATTS_EVT_RW_AUTHORIZE_REQUEST:
    NRF_LOG_DEBUG("GATT_SERVICE-- Read EVent");
    if (read_handler_ != NULL) {
      read_handler_(p_cus, p_ble_evt);
    } else {
      NRF_LOG_ERROR("GATT_SERVICE-- Read handler was NULL");
    }
    break;

  case BLE_GATTS_EVT_HVC:
    NRF_LOG_DEBUG("GATT_SERVICE-- Indicate EVent");
    if (indicate_handler_ != NULL) {
      indicate_handler_(p_cus, p_ble_evt);
    } else {
      NRF_LOG_ERROR("GATT_SERVICE-- Write handler was NULL");
    }
    break;
  default:
    // No implementation needed.
    break;
  }
}

uint32_t send_data_to_hub(ble_cus_t *p_cus, char *data) {
  NRF_LOG_DEBUG("GATT_SERVICE-- Sending data to hub of length %d %s", strlen(data), nrf_log_push(data));
  if (p_cus == NULL || data == NULL) {
    return NRF_ERROR_INVALID_PARAM;
  } // end if

  uint16_t data_len = strlen(data);
  ble_gatts_hvx_params_t hvx_params;
  memset(&hvx_params, 0, sizeof(hvx_params));
  hvx_params.handle = p_cus->custom_value_handles.value_handle;
  hvx_params.type = BLE_GATT_HVX_INDICATION;
  hvx_params.offset = 0;
  hvx_params.p_len = &data_len;
  hvx_params.p_data = (uint8_t *)data;

  return sd_ble_gatts_hvx(p_cus->conn_handle, &hvx_params);
}

uint32_t gatt_services_init(ble_cus_t *p_cus, const ble_cus_init_t *p_cus_init, on_write_handler_t write_handler, on_indicate_handler_t indicate_handler, on_read_handler_t read_handler) {
  if (p_cus == NULL || p_cus_init == NULL || write_handler == NULL || indicate_handler == NULL || read_handler == NULL) {
    NRF_LOG_ERROR("GATT_SERVICE-- Data missing or was found to be null");
    return NRF_ERROR_INVALID_PARAM;
  } // end if

  write_handler_ = write_handler;
  indicate_handler_ = indicate_handler;
  read_handler_ = read_handler;

  uint32_t err_code;
  ble_uuid_t ble_uuid;

  // Initialize service structure
  p_cus->evt_handler = p_cus_init->evt_handler;
  p_cus->conn_handle = BLE_CONN_HANDLE_INVALID;

  // Add Custom Service UUID
  ble_uuid128_t base_uuid = {BLE_SERVICE_UUID_BASE};
  err_code = sd_ble_uuid_vs_add(&base_uuid, &p_cus->uuid_type);
  VERIFY_SUCCESS(err_code);

  ble_uuid.type = p_cus->uuid_type;
  ble_uuid.uuid = DATA_TRANSFER_SERVICE_UUID;

  // Add the Custom Service
  err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY, &ble_uuid, &p_cus->service_handle);
  if (err_code != NRF_SUCCESS) {
    NRF_LOG_ERROR("GATT_SERVICE-- Error addeding the Data transfer Service");
    return err_code;
  } // end if
  NRF_LOG_DEBUG("GATT_SERVICE-- Successfully added the Data transfer Service");
  is_client_connected_ = false;
  // Add Custom Value characteristic
  return tag_add_data_characteristic(p_cus, p_cus_init);
}

void stop_gatt_service() {
  NRF_LOG_DEBUG("GATT_SERVICE-- Stopping the GATT server");
    // Disconnect all connected clients

    sd_ble_gap_disconnect(NULL, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
    is_client_connected_ = false;
    // Disable GATT server
    sd_ble_gatts_service_changed(NULL, 0x0000, 0xFFFF);
}

bool is_client_connected(){
return is_client_connected_;
}
