#include "tag_utils.h"

#include "app_button.h"
#include "app_timer.h"
#include "boards.h"

// NRF Log Libs
#include "nrf.h"
#include "nrf_drv_saadc.h"
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
#include "nrf_log_default_backends.h"

#define SAMPLES_IN_BUFFER 3
// V(P) = RESULT x REFERENCE / ( GAIN x RESOLUTION) = RESULT x (600 / (1/6 x 2^(14)) =  ADC_RESULT x 0.2197265625;
#define ADC_RESULT_IN_MILLI_VOLTS(ADC_RESULT) ((ADC_RESULT * 0.2197265625))

static uint64_t boot_time_ms = 100;

static uint32_t prev_time_ms;

static uint32_t resert_time_ms;

static uint64_t timer_offset = 0;

static uint16_t battery_level_ = 0;

static nrf_saadc_value_t m_buffer_pool[2][SAMPLES_IN_BUFFER];

static void button_event_handler(uint8_t pin_no, uint8_t button_action) {
  NRF_LOG_DEBUG("BUTTON CLICK EVENT");
  uint32_t err_code;
  switch (button_action) {
  case APP_BUTTON_PUSH:
    NRF_LOG_DEBUG("Reset button was pushed");
    break;
  case APP_BUTTON_RELEASE:
    NRF_LOG_DEBUG("Reset button was released");
    sd_nvic_SystemReset();
    break;
  default:
    return; // no implementation needed
  }
}

static app_button_cfg_t tag_buttons_cfg[] =
    {{BSP_BUTTON_0, false, BUTTON_PULL, button_event_handler}};

void tag_init_led(void) {
  NRF_LOG_DEBUG("TAG_UTILS-- Initialising the TAG Leds");
  nrf_gpio_cfg_output(TAG_GREEN_LED_PIN);
  nrf_gpio_cfg_output(TAG_RED_LED_PIN);
  nrf_gpio_cfg_output(TAG_BLUE_LED_PIN);
}

void tag_btns_init() {
  NRF_LOG_DEBUG("TAG_UTILS-- Initialising the TAG buttons");
  ret_code_t err_code;
  err_code = app_button_init(tag_buttons_cfg, ARRAY_SIZE(tag_buttons_cfg), RESET_BUTTON_DETECTION_DELAY);
  APP_ERROR_CHECK(err_code);
  err_code = app_button_enable();
  APP_ERROR_CHECK(err_code);
}

void saadc_callback(nrf_drv_saadc_evt_t const *p_event) {
  ret_code_t err_code;

  if (p_event->type == NRF_DRV_SAADC_EVT_DONE) {
    uint32_t buffer_samples = 0;
    uint16_t adc_result = 0;
    uint16_t adc_voltage_mv = 0;

    err_code = nrf_drv_saadc_buffer_convert(p_event->data.done.p_buffer, SAMPLES_IN_BUFFER);
    APP_ERROR_CHECK(err_code);

    for (uint8_t i = 0; i < SAMPLES_IN_BUFFER; i++) {
      buffer_samples += p_event->data.done.p_buffer[i];
    }

    adc_result = buffer_samples / SAMPLES_IN_BUFFER;
    adc_voltage_mv = ADC_RESULT_IN_MILLI_VOLTS(adc_result);

    NRF_LOG_INFO("voltage : %d mV.", adc_voltage_mv * 2);
    battery_level_ = adc_voltage_mv * 2;
  }
}

void saadc_init(void) {
  ret_code_t err_code;

  nrf_saadc_channel_config_t channel_config = NRF_DRV_SAADC_DEFAULT_CHANNEL_CONFIG_SE(NRF_SAADC_INPUT_AIN2);

  nrf_drv_saadc_config_t saadc_config;
  saadc_config.resolution = NRF_SAADC_RESOLUTION_14BIT;
  saadc_config.interrupt_priority = 6;
  err_code = nrf_drv_saadc_init(&saadc_config, saadc_callback);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_saadc_channel_init(0, &channel_config);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_saadc_buffer_convert(m_buffer_pool[0], SAMPLES_IN_BUFFER);
  APP_ERROR_CHECK(err_code);

  err_code = nrf_drv_saadc_buffer_convert(m_buffer_pool[1], SAMPLES_IN_BUFFER);
  APP_ERROR_CHECK(err_code);
}

void tag_init_battery_pins(void) {
  NRF_LOG_DEBUG("TAG_UTILS-- Initialising the TAG battery pins");
  nrf_gpio_cfg_input(TAG_BATTERY_CHRG_DONE_PIN, NRF_GPIO_PIN_PULLUP);
  nrf_gpio_cfg_input(TAG_BATTERY_CHARGING_PIN, NRF_GPIO_PIN_PULLUP);
  saadc_init();
  nrf_drv_saadc_sample();
}

void check_and_update_battery_state() {
  if (nrf_gpio_pin_read(TAG_BATTERY_CHRG_DONE_PIN) == 0) {
    nrf_gpio_pin_clear(TAG_GREEN_LED_PIN);
  } else if (nrf_gpio_pin_read(TAG_BATTERY_CHARGING_PIN) == 0) {
    nrf_gpio_pin_clear(TAG_RED_LED_PIN);
  } else {
    nrf_gpio_pin_set(TAG_GREEN_LED_PIN);
    nrf_gpio_pin_set(TAG_RED_LED_PIN);
  }
}

// Timer callback function
static void timer_callback(void *p_context) {
  uint32_t current_time = (app_timer_cnt_get() * 1000 / APP_TIMER_CLOCK_FREQ);
  if (prev_time_ms > current_time) {
    timer_offset += 0x1ffff;
    resert_time_ms = prev_time_ms;
  } // end if
  boot_time_ms = current_time + timer_offset;
  prev_time_ms = current_time;
  nrf_drv_saadc_sample();
}

void boot_timer_init() {
  NRF_LOG_DEBUG("TAG_UTILS-- Initialising the boot timer");

  APP_TIMER_DEF(m_repeated_timer_id);
  app_timer_init();
  uint32_t err_code;
  uint32_t ticks = APP_TIMER_TICKS(TICK_RATE_MS);
  err_code = app_timer_create(&m_repeated_timer_id, APP_TIMER_MODE_REPEATED, timer_callback);
  err_code = app_timer_start(m_repeated_timer_id, ticks, NULL);
  APP_ERROR_CHECK(err_code);
}

#ifdef TAG_DEBUG_MODE
void switch_led_debug_mode(tag_led_modes_e mode) {
  switch (mode) {
  case TAG_RECEIVING_E2_DATA:
    nrf_gpio_pin_toggle(TAG_BLUE_LED_PIN);
    nrf_delay_ms(DEBUG_LED_BLINK_DELAY_MS);
    nrf_gpio_pin_toggle(TAG_BLUE_LED_PIN);
    break;
  case TAG_TO_HUB_TRANSFER_COMPLETE:
    nrf_gpio_pin_toggle(TAG_GREEN_LED_PIN);
    nrf_delay_ms(DEBUG_LED_BLINK_DELAY_MS);
    nrf_gpio_pin_toggle(TAG_GREEN_LED_PIN);
    nrf_delay_ms(DEBUG_LED_BLINK_DELAY_MS);
    nrf_gpio_pin_toggle(TAG_GREEN_LED_PIN);
    nrf_delay_ms(DEBUG_LED_BLINK_DELAY_MS);
    nrf_gpio_pin_toggle(TAG_GREEN_LED_PIN);
    nrf_delay_ms(DEBUG_LED_BLINK_DELAY_MS);
    break;
  default:
    nrf_gpio_pin_set(TAG_GREEN_LED_PIN);
    nrf_gpio_pin_set(TAG_RED_LED_PIN);
    nrf_gpio_pin_set(TAG_BLUE_LED_PIN);
    break;
  }
}
#endif

uint16_t get_battery_level() {
  return battery_level_;
}

uint64_t millis() {
  return boot_time_ms;
}